#include "Utilities.h"

#define STB_IMAGE_IMPLEMENTATION
#include <stb_image.h>

GLuint LoadBox(GLenum* primitive, unsigned int* count)
{
    GLuint VBO;
    GLuint VAO;

    std::vector<glm::vec3> positions;
    // Front face
    positions.push_back(glm::vec3( 1.0f,  1.0f,  1.0f));
    positions.push_back(glm::vec3(-1.0f,  1.0f,  1.0f));
    positions.push_back(glm::vec3(-1.0f, -1.0f,  1.0f));
    positions.push_back(glm::vec3(-1.0f, -1.0f,  1.0f));
    positions.push_back(glm::vec3( 1.0f, -1.0f,  1.0f));
    positions.push_back(glm::vec3( 1.0f,  1.0f,  1.0f));
    // Back face
    positions.push_back(glm::vec3( 1.0f, -1.0f, -1.0f));
    positions.push_back(glm::vec3(-1.0f, -1.0f, -1.0f));
    positions.push_back(glm::vec3(-1.0f,  1.0f, -1.0f));
    positions.push_back(glm::vec3(-1.0f,  1.0f, -1.0f));
    positions.push_back(glm::vec3( 1.0f,  1.0f, -1.0f));
    positions.push_back(glm::vec3( 1.0f, -1.0f, -1.0f));
    // Top face
    positions.push_back(glm::vec3( 1.0f,  1.0f, -1.0f));
    positions.push_back(glm::vec3(-1.0f,  1.0f, -1.0f));
    positions.push_back(glm::vec3(-1.0f,  1.0f,  1.0f));
    positions.push_back(glm::vec3(-1.0f,  1.0f,  1.0f));
    positions.push_back(glm::vec3( 1.0f,  1.0f,  1.0f));
    positions.push_back(glm::vec3( 1.0f,  1.0f, -1.0f));
    // Bottom face
    positions.push_back(glm::vec3( 1.0f, -1.0f,  1.0f));
    positions.push_back(glm::vec3(-1.0f, -1.0f,  1.0f));
    positions.push_back(glm::vec3(-1.0f, -1.0f, -1.0f));
    positions.push_back(glm::vec3(-1.0f, -1.0f, -1.0f));
    positions.push_back(glm::vec3( 1.0f, -1.0f, -1.0f));
    positions.push_back(glm::vec3( 1.0f, -1.0f,  1.0f));
    // Left face
    positions.push_back(glm::vec3(-1.0f,  1.0f,  1.0f));
    positions.push_back(glm::vec3(-1.0f,  1.0f, -1.0f));
    positions.push_back(glm::vec3(-1.0f, -1.0f, -1.0f));
    positions.push_back(glm::vec3(-1.0f, -1.0f, -1.0f));
    positions.push_back(glm::vec3(-1.0f, -1.0f,  1.0f));
    positions.push_back(glm::vec3(-1.0f,  1.0f,  1.0f));
    // Right face
    positions.push_back(glm::vec3( 1.0f,  1.0f, -1.0f));
    positions.push_back(glm::vec3( 1.0f,  1.0f,  1.0f));
    positions.push_back(glm::vec3( 1.0f, -1.0f,  1.0f));
    positions.push_back(glm::vec3( 1.0f, -1.0f,  1.0f));
    positions.push_back(glm::vec3( 1.0f, -1.0f, -1.0f));
    positions.push_back(glm::vec3( 1.0f,  1.0f, -1.0f));

    std::vector<glm::vec3> normals;
    // Front face
    normals.push_back(glm::vec3(0.0f, 0.0f, 1.0f));
    normals.push_back(glm::vec3(0.0f, 0.0f, 1.0f));
    normals.push_back(glm::vec3(0.0f, 0.0f, 1.0f));
    normals.push_back(glm::vec3(0.0f, 0.0f, 1.0f));
    normals.push_back(glm::vec3(0.0f, 0.0f, 1.0f));
    normals.push_back(glm::vec3(0.0f, 0.0f, 1.0f));
    // Back face
    normals.push_back(glm::vec3(0.0f, 0.0f, -1.0f));
    normals.push_back(glm::vec3(0.0f, 0.0f, -1.0f));
    normals.push_back(glm::vec3(0.0f, 0.0f, -1.0f));
    normals.push_back(glm::vec3(0.0f, 0.0f, -1.0f));
    normals.push_back(glm::vec3(0.0f, 0.0f, -1.0f));
    normals.push_back(glm::vec3(0.0f, 0.0f, -1.0f));
    // Top face
    normals.push_back(glm::vec3(0.0f, 1.0f, 0.0f));
    normals.push_back(glm::vec3(0.0f, 1.0f, 0.0f));
    normals.push_back(glm::vec3(0.0f, 1.0f, 0.0f));
    normals.push_back(glm::vec3(0.0f, 1.0f, 0.0f));
    normals.push_back(glm::vec3(0.0f, 1.0f, 0.0f));
    normals.push_back(glm::vec3(0.0f, 1.0f, 0.0f));
    // Bottom face
    normals.push_back(glm::vec3(0.0f, -1.0f, 0.0f));
    normals.push_back(glm::vec3(0.0f, -1.0f, 0.0f));
    normals.push_back(glm::vec3(0.0f, -1.0f, 0.0f));
    normals.push_back(glm::vec3(0.0f, -1.0f, 0.0f));
    normals.push_back(glm::vec3(0.0f, -1.0f, 0.0f));
    normals.push_back(glm::vec3(0.0f, -1.0f, 0.0f));
    // Left face
    normals.push_back(glm::vec3(-1.0f, 0.0f, 0.0f));
    normals.push_back(glm::vec3(-1.0f, 0.0f, 0.0f));
    normals.push_back(glm::vec3(-1.0f, 0.0f, 0.0f));
    normals.push_back(glm::vec3(-1.0f, 0.0f, 0.0f));
    normals.push_back(glm::vec3(-1.0f, 0.0f, 0.0f));
    normals.push_back(glm::vec3(-1.0f, 0.0f, 0.0f));
    // Right face
    normals.push_back(glm::vec3(1.0f, 0.0f, 0.0f));
    normals.push_back(glm::vec3(1.0f, 0.0f, 0.0f));
    normals.push_back(glm::vec3(1.0f, 0.0f, 0.0f));
    normals.push_back(glm::vec3(1.0f, 0.0f, 0.0f));
    normals.push_back(glm::vec3(1.0f, 0.0f, 0.0f));
    normals.push_back(glm::vec3(1.0f, 0.0f, 0.0f));

    std::vector<glm::vec2> texCoord;
    // Front face
    texCoord.push_back(glm::vec2(1.0f, 1.0f));
    texCoord.push_back(glm::vec2(0.0f, 1.0f));
    texCoord.push_back(glm::vec2(0.0f, 0.0f));
    texCoord.push_back(glm::vec2(0.0f, 0.0f));
    texCoord.push_back(glm::vec2(1.0f, 0.0f));
    texCoord.push_back(glm::vec2(1.0f, 1.0f));
    // Back face
    texCoord.push_back(glm::vec2(1.0f, 1.0f));
    texCoord.push_back(glm::vec2(0.0f, 1.0f));
    texCoord.push_back(glm::vec2(0.0f, 0.0f));
    texCoord.push_back(glm::vec2(0.0f, 0.0f));
    texCoord.push_back(glm::vec2(1.0f, 0.0f));
    texCoord.push_back(glm::vec2(1.0f, 1.0f));
    // Top face
    texCoord.push_back(glm::vec2(1.0f, 1.0f));
    texCoord.push_back(glm::vec2(0.0f, 1.0f));
    texCoord.push_back(glm::vec2(0.0f, 0.0f));
    texCoord.push_back(glm::vec2(0.0f, 0.0f));
    texCoord.push_back(glm::vec2(1.0f, 0.0f));
    texCoord.push_back(glm::vec2(1.0f, 1.0f));
    // Bottom face
    texCoord.push_back(glm::vec2(1.0f, 1.0f));
    texCoord.push_back(glm::vec2(0.0f, 1.0f));
    texCoord.push_back(glm::vec2(0.0f, 0.0f));
    texCoord.push_back(glm::vec2(0.0f, 0.0f));
    texCoord.push_back(glm::vec2(1.0f, 0.0f));
    texCoord.push_back(glm::vec2(1.0f, 1.0f));
    // Left face
    texCoord.push_back(glm::vec2(1.0f, 1.0f));
    texCoord.push_back(glm::vec2(0.0f, 1.0f));
    texCoord.push_back(glm::vec2(0.0f, 0.0f));
    texCoord.push_back(glm::vec2(0.0f, 0.0f));
    texCoord.push_back(glm::vec2(1.0f, 0.0f));
    texCoord.push_back(glm::vec2(1.0f, 1.0f));
    // Right face
    texCoord.push_back(glm::vec2(1.0f, 1.0f));
    texCoord.push_back(glm::vec2(0.0f, 1.0f));
    texCoord.push_back(glm::vec2(0.0f, 0.0f));
    texCoord.push_back(glm::vec2(0.0f, 0.0f));
    texCoord.push_back(glm::vec2(1.0f, 0.0f));
    texCoord.push_back(glm::vec2(1.0f, 1.0f));

    struct Vertex
    {
        glm::vec3 position;
        glm::vec2 texCoord;
    };

    std::vector<Vertex> BufferData;

    for (int i = 0; i < positions.size(); i++)
    {
        BufferData.push_back({ positions[i], texCoord[i] });
    }

    glGenBuffers(1, &VBO);

    glBindBuffer(GL_ARRAY_BUFFER, VBO);
    glBufferData(GL_ARRAY_BUFFER, sizeof(Vertex) * BufferData.size(), BufferData.data(), GL_STATIC_DRAW);

    glGenVertexArrays(1, &VAO);
    glBindVertexArray(VAO);

    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)0);
    glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)(sizeof(glm::vec3)));

    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);

    *count = BufferData.size();
    *primitive = GL_TRIANGLES;

    return VAO;
}

GLuint LoadTexture(std::string filepath, GLint paramWrapS, GLint paramWrapT, GLint paramMinFilter, GLuint paramMagFilter)
{
    /*
        paramWrapS, paramWrapT = 
                - GL_CLAMP_TO_EDGE
                - GL_CLAMP_TO_BORDER
                - GL_MIRRORED_REPEAT
                - GL_REPEAT
                - GL_MIRROR_CLAMP_TO_EDGE

        paramMinFilter =
                - GL_NEAREST
                - GL_LINEAR
                - GL_NEAREST_MIPMAP_NEAREST
                - GL_LINEAR_MIPMAP_NEAREST
                - GL_NEAREST_MIPMAP_LINEAR
                - GL_LINEAR_MIPMAP_LINEAR
    
        paramMagFilter = 
                - GL_NEAREST
                - GL_LINEAR
    
    */

    int imageWidth;
    int imageHeight;
    int imageChannels;

    GLuint texture;

    stbi_set_flip_vertically_on_load(true);

    unsigned char* image = stbi_load(filepath.c_str(), &imageWidth, &imageHeight, &imageChannels, STBI_rgb);

    if (!image)
    {
        std::cerr << "Error loading texture  " << filepath << std::endl;

        return 0;
    }
        
    glGenTextures(1, &texture);

    glBindTexture(GL_TEXTURE_2D, texture);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, paramWrapS);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, paramWrapT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, paramMinFilter);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, paramMagFilter);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, imageWidth, imageHeight, 0, GL_RGB, GL_UNSIGNED_BYTE, image);

    if (paramMinFilter == GL_NEAREST_MIPMAP_NEAREST ||
        paramMinFilter == GL_LINEAR_MIPMAP_NEAREST ||
        paramMinFilter == GL_NEAREST_MIPMAP_LINEAR ||
        paramMinFilter == GL_LINEAR_MIPMAP_LINEAR )
    {
        glGenerateMipmap(GL_TEXTURE_2D);
    }

    glBindTexture(GL_TEXTURE_2D, 0);

    stbi_image_free(image);

    return texture;
}